/*
 * Copyright (c) 2016 Dimitris Papastamos <sin@2f30.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/socket.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <err.h>
#include <ifaddrs.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "arg.h"

char *argv0;

void
scan(char *ifname)
{
	struct ifaddrs *ifas, *ifa;
	int found = 0;

	if (getifaddrs(&ifas) == -1)
		err(1, "getifaddrs");
	if (*ifname) {
		for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
			if (!strncmp(ifname, ifa->ifa_name, IFNAMSIZ)) {
				found = 1;
				break;
			}
		}
	} else {
		for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
			if (ifa->ifa_flags & IFF_LOOPBACK)
				continue;
			if (ifa->ifa_flags & IFF_RUNNING) {
				if (ifa->ifa_flags & IFF_UP) {
					strlcpy(ifname, ifa->ifa_name, IFNAMSIZ);
					found = 1;
					break;
				}
			}
		}
	}
	freeifaddrs(ifas);
	if (!found)
		errx(1, "no usable interface found");
}

void
sample(char *ifname, unsigned long long *rxbytes, unsigned long long *txbytes,
       unsigned long long *rxpps, unsigned long long *txpps)
{
	int mib[6];
	char *buf, *next;
	size_t size;
	struct if_msghdr *ifm;
	struct sockaddr_dl *sdl = NULL;

	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = 0;
	mib[4] = NET_RT_IFLIST;
	mib[5] = 0;
	if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0)
		errx(1, "sysctl failed");
	buf = malloc(size);
	if (!buf)
		err(1, "malloc");
	if (sysctl(mib, 6, buf, &size, NULL, 0) < 0)
		errx(1, "sysctl failed");
	for (next = buf; next < buf + size; next += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)next;
		if (ifm->ifm_type != RTM_NEWADDR) {
			if (ifm->ifm_flags & IFF_UP) {
				sdl = (struct sockaddr_dl *)(ifm + 1);
				if (sdl->sdl_family != AF_LINK)
					continue;
				if (strncmp(sdl->sdl_data, ifname, sdl->sdl_nlen) != 0)
					continue;
				*rxbytes = ifm->ifm_data.ifi_ibytes;
				*txbytes = ifm->ifm_data.ifi_obytes;
				*rxpps = ifm->ifm_data.ifi_ipackets;
				*txpps = ifm->ifm_data.ifi_opackets;
				break;
			}
		}
	}
	free(buf);
	if (!sdl)
		errx(1, "interface %s cannot be sampled", ifname);
}

void
scale(char **suffix, unsigned long long *bits)
{
	static char *suffixes[] = { "bps", "Kbps", "Mbps" };

	*suffix = suffixes[0];
	if (*bits >= 1000 && *bits < 1000 * 1000) {
		*suffix = suffixes[1];
		*bits /= 1000;
	} else if (*bits >= 1000 * 1000) {
		*suffix = suffixes[2];
		*bits /= 1000 * 1000;
	}
}

void
print(char *ifname, unsigned long long rxbits, unsigned long long txbits,
      unsigned long long rxpps, unsigned long long txpps)
{
	char *rxsuffix, *txsuffix;

	scale(&rxsuffix, &rxbits);
	scale(&txsuffix, &txbits);
	printf("%s: down %lld %s up %lld %s rxpps %lld txpps %lld\n",
	       ifname, rxbits, rxsuffix, txbits, txsuffix, rxpps, txpps);
}

void
loop(char *ifname)
{
	unsigned long long oldrxbytes, rxbytes;
	unsigned long long oldtxbytes, txbytes;
	unsigned long long oldrxpps, rxpps;
	unsigned long long oldtxpps, txpps;

	sample(ifname, &oldrxbytes, &oldtxbytes, &oldrxpps, &oldtxpps);
	for (;;) {
		sample(ifname, &rxbytes, &txbytes, &rxpps, &txpps);
		print(ifname,
		      (rxbytes - oldrxbytes) * 8,
		      (txbytes - oldtxbytes) * 8,
		      rxpps - oldrxpps,
		      txpps - oldtxpps);
		oldrxbytes = rxbytes;
		oldtxbytes = txbytes;
		oldrxpps = rxpps;
		oldtxpps = txpps;
		sleep(1);
	}
}

void
usage(void)
{
	fprintf(stderr, "usage: %s [-i interface]\n", argv0);
	exit(1);
}

int
main(int argc, char *argv[])
{
	char ifname[IFNAMSIZ] = "";

	ARGBEGIN {
	case 'i':
		strlcpy(ifname, EARGF(usage()), sizeof(ifname));
		break;
	default:
		usage();
	} ARGEND

	scan(ifname);
	loop(ifname);
	return 0;
}
